import json
import re
import textwrap
from typing import Any, Mapping, Optional

import click


class ByteSizeParamType(click.ParamType):
    name = "byte"

    _rx_digits = re.compile(r'^(\d+(?:\.\d*)?)([kmgtpe]?)$', re.I)
    _scales = {
        'k': 2 ** 10,
        'm': 2 ** 20,
        'g': 2 ** 30,
        't': 2 ** 40,
        'p': 2 ** 50,
        'e': 2 ** 60,
    }

    def convert(self, value, param, ctx):
        if not isinstance(value, str):
            self.fail(f"expected string, got {value!r} of type {type(value).__name__}", param, ctx)
        m = self._rx_digits.search(value)
        if m is None:
            self.fail(f"{value!r} is not a valid byte-size expression", param, ctx)
        size = float(m.group(1))
        unit = m.group(2).lower()
        return int(size * self._scales.get(unit, 1))


class ByteSizeParamCheckType(ByteSizeParamType):
    name = "byte-check"

    def convert(self, value, param, ctx):
        if not isinstance(value, str):
            self.fail(f"expected string, got {value!r} of type {type(value).__name__}", param, ctx)
        m = self._rx_digits.search(value)
        if m is None:
            self.fail(f"{value!r} is not a valid byte-size expression", param, ctx)
        return value


def format_stats(raw_stats: Optional[str], indent='') -> str:
    if raw_stats is None:
        return "(unavailable)"
    stats = json.loads(raw_stats)
    text = "\n".join(f"- {k + ': ':18s}{v}" for k, v in stats.items())
    return "\n" + textwrap.indent(text, indent)


def format_multiline(value: Any, indent_length: int) -> str:
    buf = []
    for idx, line in enumerate(str(value).strip().splitlines()):
        if idx == 0:
            buf.append(line)
        else:
            buf.append((" " * indent_length) + line)
    return "\n".join(buf)


def format_nested_dicts(value: Mapping[str, Mapping[str, Any]]) -> str:
    """
    Format a mapping from string keys to sub-mappings.
    """
    rows = []
    if not value:
        rows.append("(empty)")
    else:
        for outer_key, outer_value in value.items():
            if isinstance(outer_value, dict):
                if outer_value:
                    rows.append(f"+ {outer_key}")
                    inner_rows = format_nested_dicts(outer_value)
                    rows.append(textwrap.indent(inner_rows, prefix="  "))
                else:
                    rows.append(f"+ {outer_key}: (empty)")
            else:
                if outer_value is None:
                    rows.append(f"- {outer_key}: (null)")
                else:
                    rows.append(f"- {outer_key}: {outer_value}")
    return "\n".join(rows)


def format_value(value: Any) -> str:
    if value is None:
        return "(null)"
    if isinstance(value, (dict, list, set)) and not value:
        return "(empty)"
    return str(value)



import json
import sys

import click
import humanize
from tabulate import tabulate

from . import admin
from ...session import Session, is_legacy_server
from ..pretty import print_error
from ..pagination import (
    get_preferred_page_size,
    echo_via_pager,
    tabulate_items,
)
from ..utils import format_nested_dicts
from ...exceptions import NoItems


def format_stats(raw_stats):

    value_formatters = {
        'bytes': lambda m: "{} / {}".format(
            humanize.naturalsize(int(m['current']), binary=True),
            humanize.naturalsize(int(m['capacity']), binary=True),
        ),
        'Celsius': lambda m: "{:,} C".format(
            float(m['current']),
        ),
        'bps': lambda m: "{}/s".format(
            humanize.naturalsize(float(m['current'])),
        ),
        'pct': lambda m: "{} %".format(
            m['pct'],
        ),
    }

    def format_value(metric):
        formatter = value_formatters.get(
            metric['unit_hint'],
            lambda m: "{} / {} {}".format(
                m['current'],
                m['capacity'],
                m['unit_hint'],
            ),
        )
        return formatter(metric)

    bufs = []
    node_metric_bufs = []
    for stat_key, metric in raw_stats['node'].items():
        if stat_key == 'cpu_util':
            num_cores = len(raw_stats['devices']['cpu_util'])
            if metric['pct'] is None:
                node_metric_bufs.append(f"{stat_key}: (calculating...) % ({num_cores} cores)")
            else:
                node_metric_bufs.append(f"{stat_key}: {metric['pct']} % ({num_cores} cores)")
        else:
            node_metric_bufs.append(f"{stat_key}: {format_value(metric)}")
    bufs.append(", ".join(node_metric_bufs))
    dev_metric_bufs = []
    for stat_key, per_dev_metric in raw_stats['devices'].items():
        dev_metric_bufs.append(f"+ {stat_key}")
        if stat_key == 'cpu_util' and len(per_dev_metric) > 8:
            dev_metric_bufs.append("  - (per-core stats hidden for large CPUs with more than 8 cores)")
        else:
            for dev_id, metric in per_dev_metric.items():
                dev_metric_bufs.append(
                    f"  - {dev_id}: {format_value(metric)}"
                )
    bufs.append("\n".join(dev_metric_bufs))
    return '\n'.join(bufs)


@admin.command()
@click.argument('agent_id')
def agent(agent_id):
    '''
    Show the information about the given agent.
    '''
    fields = [
        ('ID', 'id'),
        ('Status', 'status'),
        ('Region', 'region'),
        ('First Contact', 'first_contact'),
        ('CPU Usage (%)', 'cpu_cur_pct'),
        ('Total slots', 'available_slots'),
        ('Occupied slots', 'occupied_slots'),
        ('Hardware Metadata', 'hardware_metadata'),
        ('Live Stat', 'live_stat'),
    ]
    if is_legacy_server():
        del fields[9]
        del fields[6]
    with Session() as session:
        try:
            resp = session.Agent.detail(agent_id=agent_id,
                                        fields=(item[1] for item in fields))
        except Exception as e:
            print_error(e)
            sys.exit(1)
        rows = []
        for name, key in fields:
            if key == 'mem_cur_bytes' and resp[key] is not None:
                resp[key] = round(resp[key] / 2 ** 20, 1)
            if key in resp:
                if key == 'live_stat' and resp[key] is not None:
                    rows.append((name, format_stats(json.loads(resp[key]))))
                elif key == 'hardware_metadata':
                    rows.append((name, format_nested_dicts(json.loads(resp[key]))))
                else:
                    rows.append((name, resp[key]))
        print(tabulate(rows, headers=('Field', 'Value')))


@admin.command()
@click.option('-s', '--status', type=str, default='ALIVE',
              help='Filter agents by the given status.')
@click.option('--scaling-group', '--sgroup', type=str, default=None,
              help='Filter agents by the scaling group.')
def agents(status, scaling_group):
    '''
    List and manage agents.
    (super-admin privilege required)
    '''
    fields = [
        ('ID', 'id'),
        ('Status', 'status'),
        ('Scaling Group', 'scaling_group'),
        ('Region', 'region'),
        ('First Contact', 'first_contact'),
        ('CPU Usage (%)', 'cpu_cur_pct'),
        ('Used Memory (MiB)', 'mem_cur_bytes'),
        ('Total slots', 'available_slots'),
        ('Occupied slots', 'occupied_slots'),
    ]

    def format_item(item):
        if 'mem_cur_bytes' in item and item['mem_cur_bytes'] is not None:
            item['mem_cur_bytes'] = round(item['mem_cur_bytes'] / 2 ** 20, 1)

    try:
        if is_legacy_server():
            del fields[9]
            del fields[6]
        with Session() as session:
            page_size = get_preferred_page_size()
            try:
                items = session.Agent.paginated_list(
                    status,
                    scaling_group,
                    fields=[f[1] for f in fields],
                    page_size=page_size,
                )
                echo_via_pager(
                    tabulate_items(items, fields,
                                   item_formatter=format_item)
                )
            except NoItems:
                print("There are no matching agents.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@admin.group()
def watcher():
    '''Provides agent watcher operations.

    Watcher operations are available only for Linux distributions.
    '''


@watcher.command()
@click.argument('agent', type=str)
def status(agent):
    '''
    Get agent and watcher status.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    '''
    with Session() as session:
        try:
            status = session.AgentWatcher.get_status(agent)
            print(status)
        except Exception as e:
            print_error(e)
            sys.exit(1)


@watcher.command()
@click.argument('agent', type=str)
def agent_start(agent):
    '''
    Start agent service.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    '''
    with Session() as session:
        try:
            status = session.AgentWatcher.agent_start(agent)
            print(status)
        except Exception as e:
            print_error(e)
            sys.exit(1)


@watcher.command()
@click.argument('agent', type=str)
def agent_stop(agent):
    '''
    Stop agent service.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    '''
    with Session() as session:
        try:
            status = session.AgentWatcher.agent_stop(agent)
            print(status)
        except Exception as e:
            print_error(e)
            sys.exit(1)


@watcher.command()
@click.argument('agent', type=str)
def agent_restart(agent):
    '''
    Restart agent service.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    '''
    with Session() as session:
        try:
            status = session.AgentWatcher.agent_restart(agent)
            print(status)
        except Exception as e:
            print_error(e)
            sys.exit(1)



import textwrap
from typing import (
    AsyncIterator,
    Sequence,
)

from .base import api_function, BaseFunction
from ..request import Request
from ..pagination import generate_paginated_results

__all__ = (
    'Agent',
    'AgentWatcher',
)

_default_list_fields = (
    'id',
    'status',
    'scaling_group',
    'available_slots',
    'occupied_slots',
)

_default_detail_fields = (
    'id',
    'status',
    'scaling_group',
    'addr',
    'region',
    'first_contact',
    'cpu_cur_pct',
    'mem_cur_bytes',
    'available_slots',
    'occupied_slots',
)


class Agent(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches various agent
    information.

    .. note::

      All methods in this function class require your API access key to
      have the *admin* privilege.
    """

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        status: str = 'ALIVE',
        scaling_group: str = None,
        *,
        fields: Sequence[str] = _default_list_fields,
        page_size: int = 20,
    ) -> AsyncIterator[dict]:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        async for item in generate_paginated_results(
            'agent_list',
            {
                'status': (status, 'String'),
                'scaling_group': (scaling_group, 'String'),
            },
            fields,
            page_size=page_size,
        ):
            yield item

    @api_function
    @classmethod
    async def detail(
        cls,
        agent_id: str,
        fields: Sequence[str] = _default_detail_fields,
    ) -> Sequence[dict]:
        query = textwrap.dedent("""\
            query($agent_id: String!) {
                agent(agent_id: $agent_id) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(fields))
        variables = {'agent_id': agent_id}
        rqst = Request('POST', '/admin/graphql')
        rqst.set_json({
            'query': query,
            'variables': variables,
        })
        async with rqst.fetch() as resp:
            data = await resp.json()
            return data['agent']


class AgentWatcher(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that manipulate agent status.

    .. note::

      All methods in this function class require you to
      have the *superadmin* privilege.
    """

    @api_function
    @classmethod
    async def get_status(cls, agent_id: str) -> dict:
        """
        Get agent and watcher status.
        """
        rqst = Request('GET', '/resource/watcher')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data

    @api_function
    @classmethod
    async def agent_start(cls, agent_id: str) -> dict:
        """
        Start agent.
        """
        rqst = Request('POST', '/resource/watcher/agent/start')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data

    @api_function
    @classmethod
    async def agent_stop(cls, agent_id: str) -> dict:
        """
        Stop agent.
        """
        rqst = Request('POST', '/resource/watcher/agent/stop')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data

    @api_function
    @classmethod
    async def agent_restart(cls, agent_id: str) -> dict:
        """
        Restart agent.
        """
        rqst = Request('POST', '/resource/watcher/agent/restart')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data



import contextlib
import os
from pathlib import Path
import secrets
import signal
import subprocess
import sys
from typing import Iterator, List

import click

from . import main
from .pretty import print_info, print_fail, print_error


@contextlib.contextmanager
def container_ssh_ctx(session_ref: str, port: int) -> Iterator[Path]:
    random_id = secrets.token_hex(16)
    key_filename = "id_container"
    key_path = Path(f"~/.ssh/id_{random_id}").expanduser()
    try:
        subprocess.run(
            ["backend.ai", "download", session_ref, key_filename],
            shell=False,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError as e:
        print_fail(f"Failed to download the SSH key from the session (exit: {e.returncode}):")
        print(e.stdout.decode())
        sys.exit(1)
    os.rename(key_filename, key_path)
    try:
        print_info(f"running a temporary sshd proxy at localhost:{port} ...", file=sys.stderr)
        # proxy_proc is a background process
        proxy_proc = subprocess.Popen(
            [
                "backend.ai", "app", session_ref,
                "sshd", "-b", f"127.0.0.1:{port}",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        assert proxy_proc.stdout is not None
        lines: List[bytes] = []
        while True:
            line = proxy_proc.stdout.readline(1024)
            if not line:
                proxy_proc.wait()
                print_fail(f"Unexpected early termination of the sshd app command "
                           f"(exit: {proxy_proc.returncode}):")
                print((b"\n".join(lines)).decode())
                sys.exit(1)
            if f"127.0.0.1:{port}".encode() in line:
                break
            lines.append(line)
        lines.clear()
        yield key_path
    finally:
        proxy_proc.send_signal(signal.SIGINT)
        proxy_proc.wait()
        os.unlink(key_path)


@main.command(
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
        "allow_interspersed_args": True,
    },
)
@click.argument("session_ref",  type=str, metavar='SESSION_REF')
@click.option('-p', '--port',  type=int, metavar='PORT', default=9922,
              help="the port number for localhost")
@click.pass_context
def ssh(ctx: click.Context, session_ref: str, port: int) -> None:
    """Execute the ssh command against the target compute session

    \b
    SESSION_REF: The user-provided name or the unique ID of a running compute session.

    All remaining options and arguments not listed here are passed to the ssh command as-is.
    """
    try:
        with container_ssh_ctx(session_ref, port) as key_path:
            ssh_proc = subprocess.run(
                [
                    "ssh",
                    "-o", "StrictHostKeyChecking=no",
                    "-i", key_path,
                    "work@localhost",
                    "-p", str(port),
                    *ctx.args,
                ],
                shell=False,
                check=False,  # be transparent against the main command
            )
            sys.exit(ssh_proc.returncode)
    except Exception as e:
        print_error(e)


@main.command(
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
        "allow_interspersed_args": True,
    },
)
@click.argument("session_ref", type=str, metavar='SESSION_REF')
@click.argument("src", type=str, metavar='SRC')
@click.argument("dst", type=str, metavar='DST')
@click.option('-p', '--port',  type=str, metavar='PORT', default=9922,
              help="the port number for localhost")
@click.option('-r',  '--recursive', default=False, is_flag=True,
              help="recursive flag option to process directories")
@click.pass_context
def scp(ctx: click.Context, session_ref: str, src: str, dst: str, port: int, recursive: bool) -> None:
    """Execute the scp command against the target compute session

    \b
    The SRC and DST have the same format with the original scp command,
    either a remote path as "work@localhost:path" or a local path.

    SESSION_REF: The user-provided name or the unique ID of a running compute session.
    SRC: the source path
    DST: the destination path

    All remaining options and arguments not listed here are passed to the ssh command as-is.

    Examples:

    * Uploading a local directory to the session:

      > backend.ai scp mysess -p 9922 -r tmp/ work@localhost:tmp2/

    * Downloading a directory from the session:

      > backend.ai scp mysess -p 9922 -r work@localhost:tmp2/ tmp/
    """
    recursive_args = []
    if recursive:
        recursive_args.append("-r")
    try:
        with container_ssh_ctx(session_ref, port) as key_path:
            scp_proc = subprocess.run(
                [
                    "scp",
                    "-o", "StrictHostKeyChecking=no",
                    "-i", key_path,
                    "-P", str(port),
                    *recursive_args,
                    src, dst,
                    *ctx.args,
                ],
                shell=False,
                check=False,  # be transparent against the main command
            )
            sys.exit(scp_proc.returncode)
    except Exception as e:
        print_error(e)
